# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ RICH IMPORTS                                                                       │
# └────────────────────────────────────────────────────────────────────────────────────┘

from rich.padding import Padding
from rich.table import Table


# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ GET TABLE RENDERABLE                                                               │
# └────────────────────────────────────────────────────────────────────────────────────┘


def get_table_renderable(title, cols, rows):
    """ Returns a Rich Table constructed from supplied cols and rows """

    # Initialize Rich Table as renderable
    renderable = Table(title=title)

    # Add number column
    renderable.add_column("#")

    # Add columns
    [renderable.add_column(col) for col in cols]

    # Iterate over rows
    for i, row in enumerate(rows):

        # Add row
        renderable.add_row(str(i), *[str(r) if r else "" for r in row])

    # Pad the renderable
    renderable = Padding(renderable, (1, 1))

    # Return renderable
    return renderable


# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ GET COMMANDS RENDERABLE                                                            │
# └────────────────────────────────────────────────────────────────────────────────────┘


def get_commands_renderable(title, *commands):
    """ Returns a Rich Table containing available commands """

    # Initialize Rich Table as renderable
    renderable = Table(title=title)

    # Define columns
    cols = (
        "command",
        "arguments",
        "description",
        "example",
    )

    # Convert commands to list
    commands = [list(c) for c in commands]

    # Add quit command to end of commands
    commands = commands + [["q[uit]", "", "Quit or go back", "q"]]

    # Iterate over commands
    for i, command in enumerate(commands):

        # Get command
        command = command[0]

        # Add square bracket escape in command
        command = command.replace("[", "\[")  # noqa

        # Replace command
        commands[i][0] = command

    # Get commands renderable
    renderable = get_table_renderable(title, cols, commands)

    # Return renderable
    return renderable


# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ DISPLAY COMMANDS                                                                   │
# └────────────────────────────────────────────────────────────────────────────────────┘


def display_commands(title, console, *commands, aux_tables=None):
    """ Displays a commands renderable in a pager """

    # Initialize pager
    with console.pager():

        # Get commands renderable
        commands_renderable = get_commands_renderable(title, *commands)

        # Display commands
        console.print(commands_renderable, justify="center")

        # Check if aux tables is not null
        if aux_tables:

            # Iterate over aux tables
            for title, cols, rows in aux_tables:

                # Get renderable
                renderable = get_table_renderable(title, cols, rows)

                # Display aux table
                console.print(renderable, justify="center")



# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ GENERAL IMPORTS                                                                    │
# └────────────────────────────────────────────────────────────────────────────────────┘

import tldextract

from urllib.parse import urlparse

# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ PROJECT IMPORTS                                                                    │
# └────────────────────────────────────────────────────────────────────────────────────┘

import yank.constants as _c

from yank.request import Request


# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ TARGET                                                                             │
# └────────────────────────────────────────────────────────────────────────────────────┘


class Target:
    """ A utility class used to represent a target web page or API endpoint """

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ CLASS ATTRIBUTES                                                               │
    # └────────────────────────────────────────────────────────────────────────────────┘

    # Initialize interface to None
    interface = None

    # Initialize cached driver
    _driver = None

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ INIT METHOD                                                                    │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def __init__(self, url, yanker):
        """ Init Method """

        # Set URL
        self.url = url

        # Get parsed URL
        url_parsed = urlparse(url)

        # Set base URL
        self.url_base = self.base_url = f"{url_parsed.scheme}://{url_parsed.netloc}"

        # Set yanker
        self.yanker = yanker

        # Set requester
        self.requester = self.yanker.requester

        # Set browser
        self.browser = self.yanker.browser

        # Set has captcha
        self.has_captcha = False

        # Set captcha solved
        self.captcha_solved = False

        # Initialize requests
        self.requests = []

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ DRIVER                                                                         │
    # └────────────────────────────────────────────────────────────────────────────────┘

    @property
    def driver(self):
        """ Returns an initialized or cached Selenium webdriver instance """

        # Get browser
        browser = self.browser

        # Return None if browser is null
        if not self.browser:
            return None

        # NOTE: Driver is cached so we don't initialize a webdriver instance if the
        # initialized yanker doesn't need it

        # Check if driver is cached
        if self._driver:

            # Return cached driver
            return self._driver

        # Get driver from browser
        driver = browser.driver

        # Cache driver
        self._driver = driver

        # Return driver
        return driver

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ REQUEST                                                                        │
    # └────────────────────────────────────────────────────────────────────────────────┘

    @property
    def request(self):
        """ Returns the request that shares the target's URL """

        # Get target requests
        target_requests = [r for r in self.requests if r.url == self.url]

        # Return None if target requests is null
        if not target_requests:
            return None

        # Get target request
        target_request = target_requests[0]

        # Return target request
        return target_request

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ RESPONSE                                                                       │
    # └────────────────────────────────────────────────────────────────────────────────┘

    @property
    def response(self):
        """ Returns the response of the target request """

        # Return target request's response
        return self.request.response if self.request else None

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ STATUS CODE                                                                    │
    # └────────────────────────────────────────────────────────────────────────────────┘

    @property
    def status_code(self):
        """ Returns the status code of the request that shares the target's URL """

        # Return target request status code
        return self.response.status_code if self.response else None

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ HTML                                                                           │
    # └────────────────────────────────────────────────────────────────────────────────┘

    @property
    def html(self):
        """ Returns the HTML of the target's response object """

        # Return HTML
        return self.response.html

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ JSON                                                                           │
    # └────────────────────────────────────────────────────────────────────────────────┘

    @property
    def json(self):
        """ Returns a JSON dict of the target's response object """

        # Return JSON
        return self.response.json

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ SOUP                                                                           │
    # └────────────────────────────────────────────────────────────────────────────────┘

    @property
    def soup(self):
        """ Returns a HTML BeautifulSoup of the target's response object """

        # Return soup
        return self.response.soup

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ GET                                                                            │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def get(self, driver_callback=None, solve_captcha_callback=None):
        """ Performs an HTTP GET request to the page using its yanker's requester """

        # Get URL
        url = self.url

        # Get yanker
        yanker = self.yanker

        # Determine if should get auto headers
        should_get_auto_headers = yanker.auto_headers and yanker._auto_headers is None

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ DRIVER                                                                     │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Determine if should use driver
        should_use_driver = self.browser and (
            driver_callback or solve_captcha_callback or should_get_auto_headers
        )

        # Check if driver is not null
        if should_use_driver:

            # Get driver
            driver = self.driver

            # TODO: Copy cookies over from session

            # Check if driver callback is not null
            if driver_callback:

                # Check if driver mode is not none and callback has stop when decorator
                if getattr(driver_callback, "has_stop_when", False):

                    # Get quick driver
                    driver_quick = self.browser.driver_quick

                    # Copy cookies from driver to quick driver

                    # Get URL with quick driver
                    driver_quick.get(url)

                    # Copy cookies from quick driver back to driver

                    # Set local driver to quick driver
                    driver = driver_quick

                # Otherwise get URL as usual
                else:

                    # Get URL with driver
                    driver.get(url)

                # Execute driver callback
                driver_callback(driver)

            # Otherwise handle case of no driver callback
            else:

                # Get URL with driver
                driver.get(url)

            # Check if solve captcha callback is not null
            if solve_captcha_callback:

                # TODO: HOW TO MARK AS SOLVED?

                # Pass driver into solve captcha callback
                solve_captcha_callback(driver)

            # Iterate over driver requests
            for request in driver.requests:

                # Get response
                response = request.response

                # Check if response is not None
                if response is not None:

                    # Set request of response
                    response.request = request

                # Initialize request object
                request = Request(url, driver=driver)

                # Set request response
                request.set_response(response)

                # Append request to requests
                self.requests.append(request)

                # ┌────────────────────────────────────────────────────────────────────┐
                # │ AUTO HEADERS                                                       │
                # └────────────────────────────────────────────────────────────────────┘

                # Check if should get auto headers
                if should_get_auto_headers and yanker._auto_headers is None:

                    # Extract registered domain from request URL
                    domain = tldextract.extract(request.url).registered_domain

                    # Check if registered domain is in the target URL
                    if domain in url:

                        # Get auto headers
                        _auto_headers = request.headers

                        # Update auto headers by default headers
                        _auto_headers.update(yanker.default_headers)

                        # Set auto headers cache to request headers
                        yanker._auto_headers = request.headers

                        # Set default headers
                        yanker.default_headers = _auto_headers

                        # Break here
                        break

            # Check if mode is session
            if self.yanker.mode == _c.SESSION:

                # Get driver user agent
                driver_user_agent = driver.execute_script("return navigator.userAgent;")

                # Update session user agent
                self.requester.headers.update({"User-Agent": driver_user_agent})

                # Get driver cookies
                driver_cookies = driver.get_cookies()

                # Update session cookies
                self.requester.cookies.update(
                    {c["name"]: c["value"] for c in driver_cookies}
                )

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ REQUESTER                                                                  │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Otherwise handle no driver
        else:

            # Initialize request object
            request = Request(url)

            # Append request to requests
            self.requests.append(request)

            # Initialize request kwargs
            request_kwargs = {}

            # Check if yanker mode is transient
            if self.yanker.mode == _c.TRANSIENT:

                # TODO: Implement a dynamic get headers method for users to customize

                # Add default headers to request kwargs
                request_kwargs[_c.HEADERS] = self.yanker.default_headers

            # Make an HTTP request
            response = self.requester.get(url, **request_kwargs)

            # Set request response
            request.set_response(response)

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ FILTER_REQUESTS                                                                │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def filter_requests(self, **kwargs):
        """ Filters the target's requests by given keyword arguments """

        # Get requests
        requests = self.requests

        # Filter requests
        requests = [
            r
            for r in requests
            if all([getattr(r, k, None) == v for k, v in kwargs.items()])
        ]

        # Return requests
        return requests



# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ GENERAL IMPORTS                                                                    │
# └────────────────────────────────────────────────────────────────────────────────────┘

import re

# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ RICH IMPORTS                                                                       │
# └────────────────────────────────────────────────────────────────────────────────────┘

from rich.console import Console
from rich.padding import Padding
from rich.table import Table

# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ PROJECT IMPORTS                                                                    │
# └────────────────────────────────────────────────────────────────────────────────────┘

import yank.constants as _c

from yank.tools import display_commands


# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ YANKER DISPLAY MIXIN                                                               │
# └────────────────────────────────────────────────────────────────────────────────────┘


class YankerDisplayMixin:
    """ Yanker Display Mixin """

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ CLASS ATTRIBUTES                                                               │
    # └────────────────────────────────────────────────────────────────────────────────┘

    # Initialize cached console to None
    _console = None

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ TABLE NAMES                                                                    │
    # └────────────────────────────────────────────────────────────────────────────────┘

    @property
    def table_names(self):
        """ Returns a list of table names in the database """

        # Return table names
        return list(self.tables.keys())

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ CONSOLE                                                                        │
    # └────────────────────────────────────────────────────────────────────────────────┘

    @property
    def console(self):
        """ Returns a cached or newly initialized Rich console object """

        # Check if console is cached
        if self._console:

            # Return cached console
            return self._console

        # Initialize a new Rich console
        console = Console()

        # Cache console
        self._console = console

        # Return console
        return console

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ DISPLAY TABLES                                                                 │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def display_tables(self, interactive=False):
        """ Displays a list of tables using a Rich Table """

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ STATIC                                                                     │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Get tables
        tables = [(name, interface) for name, interface in self.tables.items()]

        # Get console
        console = self.console

        # Clear console if interactive
        interactive and console.clear()

        # Get renderable
        renderable = self.get_tables_renderable(tables=tables, interactive=interactive)

        # Print renderable
        console.print(renderable, justify="center")

        # Return if not interactive
        if not interactive:
            return

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ INTERACTIVE                                                                │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Initialize while loop
        while True:

            # Get command
            command = console.input(_c.INPUT_TAG)

            # Handle case of quit
            if command in ["q", "quit"]:
                break

            # Otherwise handle case of command
            elif command in ["c", "command", "commands"]:

                # Display commands
                display_commands(
                    "Available Commands",
                    console,
                    ("r[ows]", "<#>", "Show table rows", "r 0"),
                    ("c[ols]", "<#>", "Show table columns", "c 0"),
                )

            # Get command and index
            match = re.search(r"(\w+) *(\d+)", command)

            # Check if match is not null
            if match:

                # Separate index from command
                command, idx = (match.group(1).lower().strip(), int(match.group(2)))

                # Initialize try-except block
                try:

                    # Get table name and interface
                    table_name, interface = tables[idx]

                # Handle IndexError
                except IndexError:

                    # Set table name and interface to None
                    table_name = interface = None

                # Check if table name and interface are not null
                if table_name and interface:

                    # Handle case of columns
                    if command in ["c", "cols", "columns"]:

                        # Display table
                        self.display_table(table_name, interactive=interactive)

                    # Otherwise handle case of rows
                    elif command in ["r", "rows"]:

                        # Display interface list view
                        interface.display_list(interactive=interactive)

            # Clear console
            console.clear()

            # Print renderable
            console.print(renderable, justify="center")

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ GET TABLES RENDERABLE                                                          │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def get_tables_renderable(self, tables, interactive=False):
        """ Displays a list of tables using a Rich Table """

        # Get table count
        table_count = len(tables)

        # Initialize Rich Table as renderable
        renderable = Table(
            title=(
                f"Tables: {self.db_name} "
                f"({table_count} {self.inflector.plural('table', table_count)})"
            )
        )

        # Check if interactive
        if interactive:

            # Set caption
            renderable.caption = "c\[ommands]: Show available commands"  # noqa

        # Add columns
        [
            renderable.add_column(col_name)
            for col_name in ("#", "name", "interface", "cols", "rows")
        ]

        # Iterate over tables
        for i, (table_name, interface) in enumerate(tables):

            # Get column count
            col_count = str(len(interface.Item.__table__.columns))

            # Get row count
            row_count = str(interface.count())

            # Add row
            renderable.add_row(
                str(i), interface.db_table_name, interface.name, col_count, row_count
            )

        # Pad the renderable
        renderable = Padding(renderable, (1, 1))

        # Return renderable
        return renderable

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ DISPLAY TABLE                                                                  │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def display_table(self, name, interactive=False):
        """ Displays a list of columns of a table using a Rich Table """

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ STATIC                                                                     │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Get table renderable
        renderable = self.get_table_renderable(name, interactive=interactive)

        # Get console
        console = self.console

        # Clear console if interactive
        interactive and console.clear()

        # Print renderable
        console.print(renderable, justify="center")

        # Return if not interactive
        if not interactive:
            return

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ INTERACTIVE                                                                │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Initialize while loop
        while True:

            # Get command
            command = console.input(_c.INPUT_TAG)

            # Handle case of quit
            if command in ["q", "quit"]:
                break

            # Otherwise handle case of command
            elif command in ["c", "command", "commands"]:

                # Display commands
                display_commands(title="Available Commands", console=console)

            # Clear console
            console.clear()

            # Print renderable
            console.print(renderable, justify="center")

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ GET TABLE RENDERABLE                                                           │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def get_table_renderable(self, name, interactive=False):
        """ Returns a Rich table renderable """

        # Get interface
        interface = self.tables.get(name)

        # Return if interface is null
        if not interface:
            return

        # Get field map
        field_map = interface.field_map

        # Get column count
        col_count = len(field_map)

        # Initialize Rich Table as renderable
        renderable = Table(
            title=f"Table: {name} ({col_count} "
            f"{self.inflector.plural('col', col_count)})"
        )

        # Check if interactive
        if interactive:

            # Set caption
            renderable.caption = "c\[ommands]: Show available commands"  # noqa

        # Add columns
        [
            renderable.add_column(col_name)
            for col_name in ("#", "field", "display", "cast", "type")
        ]

        # Iterate over field map
        for i, (field, info) in enumerate(field_map.items()):

            # Get cast
            cast = info[_c.CAST]

            # Add row
            renderable.add_row(
                str(i),
                field,
                info[interface.DISPLAY],
                str(cast),
                str(interface.TYPE_MAP[cast]),
            )

        # Pad the renderable
        renderable = Padding(renderable, (1, 1))

        # Return renderable
        return renderable



# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ GENERAL IMPORTS                                                                    │
# └────────────────────────────────────────────────────────────────────────────────────┘

import os

# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ SELENIUM IMPORTS                                                                   │
# └────────────────────────────────────────────────────────────────────────────────────┘

from selenium import webdriver
from seleniumwire import webdriver as webdriver_wire
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
from selenium.webdriver.firefox.options import Options as FirefoxOptions
from selenium.webdriver.support.ui import WebDriverWait
from webdriver_manager.chrome import ChromeDriverManager
from webdriver_manager.firefox import GeckoDriverManager
from webdriver_manager.utils import ChromeType

# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ PROJECT IMPORTS                                                                    │
# └────────────────────────────────────────────────────────────────────────────────────┘

import yank.constants as _c

from yank.exceptions import UnsupportedBrowserError, UnsupportedDriverModeError


# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ BROWSER                                                                            │
# └────────────────────────────────────────────────────────────────────────────────────┘


class Browser:
    """ A utility class used to represent a web browser """

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ CONSTANTS                                                                      │
    # └────────────────────────────────────────────────────────────────────────────────┘

    # Driver modes
    EAGER = _c.EAGER
    NORMAL = _c.NORMAL
    QUICK = _c.QUICK

    # Define driver modes
    DRIVER_MODES = [EAGER, QUICK, NORMAL]

    # Browsers
    CHROME = _c.CHROME
    CHROMIUM = _c.CHROMIUM
    FIREFOX = _c.FIREFOX

    # Define supported browsers
    SUPPORTED_BROWSERS = {CHROME: "Chrome", CHROMIUM: "Chromium", FIREFOX: "Firefox"}

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ INIT METHOD                                                                    │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def __init__(
        self, slug, driver_mode=NORMAL, driver_headless=False, driver_requests=False
    ):
        """ Init Method """

        # Get driver modes
        driver_modes = self.DRIVER_MODES

        # Check if driver mode not in driver modes
        if driver_mode not in driver_modes:

            # Raise UnsupportedDriverModeError
            raise UnsupportedDriverModeError(
                f"Driver mode '{driver_mode}' not supported. Please use one of the "
                f"following: {', '.join(driver_modes)}"
            )

        # Get supported browsers
        supported_browsers = self.SUPPORTED_BROWSERS

        # Check if slug not in supported browsers
        if slug not in supported_browsers:

            # Raise UnsupportedBrowserError
            raise UnsupportedBrowserError(
                f"Browser '{slug}' not supported. Please use one of the "
                f"following: {', '.join(list(supported_browsers))}"
            )

        # Set name
        self.name = supported_browsers[slug]

        # Set slug
        self.slug = slug

        # Set driver mode
        self.driver_mode = driver_mode

        # Set driver headless boolean
        self.driver_headless = driver_headless

        # Set driver requests boolean
        self.driver_requests = driver_requests

        # Initialize driver cache
        self._driver = None

        # Initialize quick driver cache
        self._driver_quick = None

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ DRIVER                                                                         │
    # └────────────────────────────────────────────────────────────────────────────────┘

    @property
    def driver(self):
        """ Returns an initialized or cached Selenium webdriver instance """

        # NOTE: Driver is cached so we don't initialize a webdriver instance if the
        # initialized yanker doesn't need it

        # Check if driver is cached
        if self._driver:

            # Return cached driver
            return self._driver

        # Initialize a new driver
        driver = self.initialize_driver(
            self.slug, self.driver_mode, self.driver_headless
        )

        # Cache driver
        self._driver = driver

        # Return driver
        return driver

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ DRIVER QUICK                                                                   │
    # └────────────────────────────────────────────────────────────────────────────────┘

    @property
    def driver_quick(self):
        """
        Returns an initialized Selenium webdriver instance with no page load strategy
        """

        # Get constants
        QUICK = self.QUICK

        # Check if driver mode is quick
        if self.driver_mode == QUICK:

            # Return the driver
            return self.driver

        # Check if the quick driver is cached
        if self._driver_quick:

            # Return the cached driver
            return self._driver_quick

        # Initialize a new quick driver
        driver_quick = self.initialize_driver(self.slug, _c.NONE, self.driver_headless)

        # Set quick driver
        self._driver_quick = driver_quick

        # Return quick driver
        return driver_quick

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ INITIALIZE DRIVER                                                              │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def initialize_driver(self, slug, driver_mode, driver_headless):
        """ Initializes a Selenium webdriver instance based on the browser slug """

        # Set log level to 0
        # Silences webdriver manager log output
        os.environ["WDM_LOG_LEVEL"] = "0"

        # Define desired capability args
        desired_capability_args = {"pageLoadStrategy": driver_mode}

        # Check if browser is Firefox
        if slug == _c.FIREFOX:

            # Initialize options
            options = FirefoxOptions()

            # Add headless option
            options.headless = driver_headless

            # Initialize desired capabilities
            desired_capabilities = DesiredCapabilities.FIREFOX

            # Set page load strategy to None
            desired_capabilities.update(desired_capability_args)

            # Handle case of Selenium Wire (driver_requests == True)
            if self.driver_requests:

                # Initialize a Firefox driver
                driver = webdriver_wire.Firefox(
                    executable_path=GeckoDriverManager().install(),
                    options=options,
                    desired_capabilities=desired_capabilities,
                )

            # Otherwise handle case of normal webdriver
            else:

                # Initialize a Firefox driver
                driver = webdriver.Firefox(
                    executable_path=GeckoDriverManager().install(),
                    options=options,
                    desired_capabilities=desired_capabilities,
                )

        # Otherwise handle default case
        else:

            # Initialize options
            options = ChromeOptions()

            # Add headless option
            options.headless = driver_headless

            # Disable blink features
            # Helps to evade CloudFlare asking you to prove you are human with captcha
            options.add_argument("--disable-blink-features=AutomationControlled")

            # See https://stackoverflow.com/questions/64165726/
            # selenium-stuck-on-checking-your-browser-before-accessing-url

            # Initialize desired capabilities
            desired_capabilities = DesiredCapabilities.CHROME

            # Set page load strategy to None
            desired_capabilities.update(desired_capability_args)

            # Define Chrome driver kwargs
            driver_kwargs = {}

            # Check if browser is Chromium
            if slug == _c.CHROMIUM:

                # Add Chrome type to kwargs
                driver_kwargs["chrome_type"] = ChromeType.CHROMIUM

            # Handle case of Selenium Wire (driver_requests == True)
            if self.driver_requests:

                # Initialize a Chrome driver
                driver = webdriver_wire.Chrome(
                    executable_path=ChromeDriverManager().install(),
                    options=options,
                    desired_capabilities=desired_capabilities,
                    **driver_kwargs,
                )

            # Otherwise handle case of normal webdriver
            else:

                # Initialize a Chrome driver
                driver = webdriver.Chrome(
                    executable_path=ChromeDriverManager().install(),
                    options=options,
                    desired_capabilities=desired_capabilities,
                    **driver_kwargs,
                )

                # Initialize driver requests to an empty list
                driver.requests = []

            # NOTE: We only use Selenium Wire when we want to capture HTTP requests
            # because Selenium Wire has a higher chance of getting blocked in some cases

            # Check if headless is True
            if driver_headless:

                # Get default user agent
                user_agent = driver.execute_script("return navigator.userAgent;")

                # Remove indication that user agent is a headless browser
                user_agent = user_agent.replace("HeadlessChrome", "Chrome")

                # Replace driver user agent
                driver.execute_cdp_cmd(
                    "Network.setUserAgentOverride", {"userAgent": user_agent}
                )

        # Return driver
        return driver

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ STOP WHEN                                                                      │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def stop_when(condition, timeout=10):
        """ Stops a driver from loading further when a certain condition is met """

        # Define decorator
        def decorator(method):

            # Define wrapper
            def wrapper(instance, driver):

                # Wait for listing nav links to appear
                WebDriverWait(driver, timeout).until(condition)

                # Stop further loading
                driver.execute_script("window.stop();")

                # Execute original method
                return method(instance, driver)

            # Set has stop when attribute on wrapped method
            wrapper.has_stop_when = True

            # Return wrapper
            return wrapper

        # Return the decorator
        return decorator



# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ GENERAL IMPORTS                                                                    │
# └────────────────────────────────────────────────────────────────────────────────────┘

import copy
import inflect
import inspect
import random
import re
import requests
import time

# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ SQL ALCHEMY IMPORTS                                                                │
# └────────────────────────────────────────────────────────────────────────────────────┘

from sqlalchemy import create_engine, event
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker


# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ PROJECT IMPORTS                                                                    │
# └────────────────────────────────────────────────────────────────────────────────────┘

import yank.constants as _c

from yank.browser import Browser
from yank.exceptions import SessionLimitReached
from yank.interface import Interface
from yank.yanker_display_mixin import YankerDisplayMixin
from yank.yanker_util_mixin import YankerUtilMixin


# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ INFLECTOR                                                                          │
# └────────────────────────────────────────────────────────────────────────────────────┘

# Initialize inflector for singular / plural conversions
inflector = inflect.engine()

# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ DATABASE                                                                           │
# └────────────────────────────────────────────────────────────────────────────────────┘

# Initialize base class
DBBase = declarative_base()


# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ YANKER                                                                             │
# └────────────────────────────────────────────────────────────────────────────────────┘


class Yanker(YankerDisplayMixin, YankerUtilMixin):
    """ A base class for custom Yanker classes """

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ PENDING FEATURES                                                               │
    # └────────────────────────────────────────────────────────────────────────────────┘

    # TODO: Implement a dynamic get headers method for users to customize
    # TODO: Enforce transience in the case of driver (new driver each time)
    # TODO: Implement session transfer from requester to driver (not done yet)
    # TODO: Auto set driver mode to quick if all yank methods have a driver method with
    #       a stop when decorator
    # TODO: Allow for sharing of single interface between methods
    # TODO: Implement skip by URL as a standalone decorator
    # TODO: Truncate list view if row count is huge
    # TODO: Improve list view reset to be able to choose what to reset

    # TODO: Improve session limits with non-items
    # TODO: Improve uniqueness handling
    # TODO: Build a CSV / XLSX exporter
    # TODO: Figure out how to share an interface between Yankers or Merger
    # TODO: Implement a hardcoded auto header as first try
    # TODO: Get base URL from current target
    # TODO: Fix missing taget.response when driver requests is False
    # TODO: Implement batches as a column in database for freelance work
    # TODO: Implement ability to not define start URLs
    # TODO: Add ID to columns in display table
    # TODO: Do not initialize database if not interface / (history)?
    # TODO: Implement generic captcha detector
    # TODO: Accept return statement in yank methods

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ CONSTANTS                                                                      │
    # └────────────────────────────────────────────────────────────────────────────────┘

    # Modes
    SESSION = _c.SESSION
    TRANSIENT = _c.TRANSIENT

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ CLASS ATTRIBUTES                                                               │
    # └────────────────────────────────────────────────────────────────────────────────┘

    # Define requester
    requester = requests

    # Initialize Browser class so that users can easily access its constants
    Browser = Browser

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ CUSTOMIZABLE CLASS ATTRIBUTES                                                  │
    # └────────────────────────────────────────────────────────────────────────────────┘

    # Initialize mode to transient
    mode = TRANSIENT

    # Initialize start URL to empty string
    start_url = ""

    # Initialize start URLs to None
    start_urls = None

    # Initialize throttle to None
    throttle_ms = None

    # Initialize auto headers to False
    auto_headers = False

    # Initialize default headers to None
    default_headers = None

    # Initialize default browser
    default_browser = Browser.CHROME

    # Initialize driver mode to normal
    driver_mode = Browser.NORMAL

    # Initialize driver headless to False
    driver_headless = False

    # Initialize driver requests to False
    driver_requests = False

    # Initialize database name to None
    db_name = None

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ INIT METHOD                                                                    │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def __init__(
        self,
        start_url=None,
        start_urls=None,
        mode="",
        auto_headers=None,
        default_headers=None,
        default_browser="",
        driver_mode=None,
        driver_headless=None,
        driver_requests=None,
        db_name=None,
    ):
        """ Init Method """

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ INFLECTOR                                                                  │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Set inflector
        self.inflector = inflector

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ MODE                                                                       │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Set mode
        self.mode = mode if mode else self.mode

        # Check if mode is session
        if self.mode == _c.SESSION:

            # Set requester to a new requests session
            self.requester = requests.Session()

            # Set default headers
            self.requester.headers.update(self.default_headers)

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ DATABASE                                                                   │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Initialize tables to empty dict
        self.tables = {}

        # Set database name
        self.db_name = db_name or self.db_name

        # Check if database name is None
        if self.db_name is None:

            # Get class name
            class_name = self.__class__.__name__

            # Set database name as snake case version of class name
            self.db_name = re.sub(r"(?<!^)(?=[A-Z])", "_", class_name).lower()

        # Initialize database engine
        self.db_engine = create_engine(f"sqlite:///{self.db_name}.db")

        # Setup database engine event listener with SQLite
        @event.listens_for(self.db_engine, "connect")
        def listener_sqlite(dbapi_connection, connection_record):

            # Enable case sensitive like / contains
            dbapi_connection.execute("pragma case_sensitive_like=ON")

            # Register exponent function
            dbapi_connection.create_function("pow", 2, lambda x, y: x ** y)

        # Create tables
        DBBase.metadata.create_all(self.db_engine)

        # Make a database session class
        DBSession = sessionmaker(bind=self.db_engine)

        # Initialize database session
        self.db_session = DBSession()

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ START URLS                                                                 │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Initialize start URLs from class attribute
        self.start_urls = start_urls or self.start_urls or []

        # Check if start URLs is a string
        if type(self.start_urls) is str:

            # Convert to list
            self.start_urls = [self.start_urls]

        # Initialize start URL
        self.start_url = start_url or self.start_url

        # Get start URL
        start_url = self.start_url

        # Check if start URL is defined
        if start_url:

            # Convert to list
            start_url = (
                list(start_url) if type(start_url) in [list, tuple] else [start_url]
            )

            # Add start URL to start URLs
            self.start_urls = start_url + self.start_urls

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ HEADERS                                                                    │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Set auto headers
        self.auto_headers = (
            auto_headers if auto_headers is not None else self.auto_headers
        )

        # Initialize auto headers cache
        self._auto_headers = None

        # Get default headers
        default_headers = self.default_headers

        # Initialize default headers
        self.default_headers = copy.deepcopy(default_headers) if default_headers else {}

        # Check if additional default headers were passed in
        if default_headers:

            # Update the instance's default headers
            self.default_headers.update(default_headers)

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ WRAP METHODS                                                               │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Get default browser and supported browsers
        default_browser = default_browser or self.default_browser

        # Call wrap methods
        # This will determine whether the user has defined a driver callback method
        has_driver_callback, has_solve_captcha_callback = self.wrap_methods(
            default_browser=default_browser
        )

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ BROWSER AND DRIVER                                                         │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Set browser to None
        self.browser = None

        # Determine if Selenium driver is required
        driver_is_required = (
            self.auto_headers or has_driver_callback or has_solve_captcha_callback
        )

        # Check if driver is required
        if driver_is_required:

            # Get driver mode
            driver_mode = driver_mode if driver_mode is not None else self.driver_mode

            # Get driver headless
            driver_headless = (
                driver_headless if driver_headless is not None else self.driver_headless
            )

            # Get driver requests
            driver_requests = (
                driver_requests if driver_requests is not None else self.driver_requests
            )

            # Initialize and set browser
            self.browser = Browser(
                default_browser,
                driver_mode=driver_mode,
                driver_headless=driver_headless,
                driver_requests=driver_requests,
            )

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ _YANK                                                                          │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def _yank(self, target=None, **kwargs):
        """ Runs the yanker on its start URLs """

        # Check if target URL is defined
        if target:

            # Call yank start on target URL and return
            self.yank_start(target, **kwargs)
            return

            # NOTE: We do this in case the user calls their initial yank method again
            # which has since been renamed to yank start

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ ITERATE OVER START URLS                                                    │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Iterate over start URLs
        for start_url in self.start_urls:

            # Initialize try-except block
            try:

                # Call yank method on start URL
                self.yank_start(start_url)

                # NOTE: yank_start is actually yank but renamed upon being wrapped
                # This is so that _yank can be renamed to yank, which will allow the
                # user to automatigically call Yanker.yank() without any arguments

            # Except graceful exceptions
            except (KeyboardInterrupt, SessionLimitReached):

                # Pass onto end of method
                pass

            # Except any exception
            except Exception:

                # Close driver to avoid lingering headless browser instances
                self.close_driver()

                # Re-raise the exception
                raise

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ CLOSE DRIVER                                                               │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Close driver
        self.close_driver()

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ YANK                                                                           │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def yank(self, target):
        """ The action performed on each of the yanker's start URLs """

        # Raise NotImplementedError
        raise NotImplementedError

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ READ                                                                           │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def read(self):
        """ Displays an interactive console of tables in the database """

        # Display tables in interactive mode
        self.display_tables(interactive=True)

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ INTERFACE                                                                      │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def interface(**kwargs):
        """
        Registers an interface with a yank method
        This effectively marks the method's return value to be commited to the daatabase
        """

        # Define decorator
        def decorator(method):

            # Get method name
            method_name = method.__name__

            # Get database table name from method name
            db_table_name = method_name.replace("yank_", "")

            # Convert database table name to singular form
            db_table_name = inflector.singular_noun(db_table_name) or db_table_name

            # Convert database table name to lower- snake-case
            db_table_name = db_table_name.lower().replace(" ", "")

            # Rename in case of yank start method
            db_table_name = "item" if db_table_name == "yank" else db_table_name

            # Initialize an interface
            interface = Interface(
                DBBase, db_table_name=db_table_name, inflector=inflector, **kwargs
            )

            # Define wrapper
            def wrapper(instance, target, *args, **kwargs):

                # Set interface on target
                target.interface = interface

                # Execute original method
                return method(instance, target, *args, **kwargs)

            # Add the interface as an attribute on the wrapped method
            wrapper.interface = interface

            # Return wrapper
            return wrapper

        # Return the decorator
        return decorator

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ WRAP METHODS                                                                   │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def wrap_methods(self, default_browser):
        """ Wraps user-defined methods in existing Yanker utility methods """

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ WRAPPER                                                                    │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Define a method wrapper
        def wrapper(
            method,
            driver_callback,
            clean_callback,
            has_captcha_callback,
            solve_captcha_callback,
        ):
            """ Wraps a yank method to handle user-defined class logic """

            # ┌────────────────────────────────────────────────────────────────────────┐
            # │ GET INTERFACE                                                          │
            # └────────────────────────────────────────────────────────────────────────┘

            # Get interface from method if it exists
            interface = getattr(method, "interface", None)

            # Check if interface is not null
            if interface:

                # Set default browser
                interface.default_browser

                # Add instance session to interface
                interface.db_session = self.db_session

                # Get database table name
                db_table_name = interface.db_table_name

                # Convert database table name to Pascal Case
                interface_name = "".join(
                    [
                        w.title() if len(w) > 2 else w.upper()
                        for w in db_table_name.split("_")
                    ]
                )

                # Set interface name
                interface.name = interface_name

                # Add interface as a "table" accessible on the yanker instance
                setattr(self, interface_name, interface)

                # Add interface to tables dict
                self.tables[db_table_name] = interface

            # Define wrapped method
            def wrapped(target, *args, **kwargs):

                # ┌────────────────────────────────────────────────────────────────────┐
                # │ PRE-REQUEST FILTERS                                                │
                # └────────────────────────────────────────────────────────────────────┘

                # Get interface from method if it exists
                interface = getattr(method, "interface", None)

                # Check if interface is not null
                if interface:

                    # ┌────────────────────────────────────────────────────────────────┐
                    # │ SESSION LIMIT                                                  │
                    # └────────────────────────────────────────────────────────────────┘

                    # Get interface session limit
                    session_limit = interface.session_limit

                    # Check if interface session count has reached session limit
                    if session_limit and interface.session_count >= session_limit:

                        # Raise SessionLimitReached
                        raise SessionLimitReached

                    # ┌────────────────────────────────────────────────────────────────┐
                    # │ SKIP BY URL                                                    │
                    # └────────────────────────────────────────────────────────────────┘

                    # Get skip by URL boolean
                    skip_by_url = interface.skip_by_url

                    # Check if should skip by URL
                    if skip_by_url:

                        # Return None if item with target URL exists
                        if interface.exists(url=target):
                            return None

                # Otherwise handle case of method decorator
                else:

                    pass

                # ┌────────────────────────────────────────────────────────────────────┐
                # │ MAKE REQUEST                                                       │
                # └────────────────────────────────────────────────────────────────────┘

                # Get throttle ms
                throttle_ms = self.throttle_ms

                # Check if throttle ms is a list or tuple
                if type(throttle_ms) in [list, tuple]:

                    # Generate a random integer between min and max
                    throttle_ms = random.randint(min(throttle_ms), max(throttle_ms))

                # Check if throttle ms is not null
                if throttle_ms:

                    # Implement sleep to throttle the request
                    time.sleep(throttle_ms / 1000)

                # Get target object from tarket URL
                target = self.get(target, driver_callback=driver_callback)

                # Get status code
                status_code = target.status_code

                # Get status color
                status_color = "green" if 200 <= status_code < 300 else "red"

                # Construct log
                log = f"GET {target.url} [bold {status_color}]{status_code}[/]"

                # Log request
                self.console.log(log)

                # Check if has captcha callback
                if has_captcha_callback:

                    # Check if target has captcha
                    if has_captcha_callback(target):

                        # Set target has captcha
                        target.has_captcha = True

                        # Set target captcha solved to False
                        target.captcha_solved = False

                        # Check if solve captcha callback
                        if solve_captcha_callback:

                            # Get the target with captcha callback
                            target = self.get(
                                target.url,
                                driver_callback=driver_callback,
                                solve_captcha_callback=solve_captcha_callback,
                            )

                            # Get target object from tarket URL
                            target = self.get(
                                target.url, driver_callback=driver_callback
                            )

                # Iterate over generated items
                for item in method(target, *args, *kwargs):

                    # ┌────────────────────────────────────────────────────────────────┐
                    # │ CLEAN RESULT                                                   │
                    # └────────────────────────────────────────────────────────────────┘

                    # Check if clean callback exists
                    if clean_callback:

                        # Pass result through clean callback
                        item = clean_callback(target.interface, item) or item

                    # ┌────────────────────────────────────────────────────────────────┐
                    # │ STORE ITEM                                                     │
                    # └────────────────────────────────────────────────────────────────┘

                    # Get interface
                    interface = target.interface

                    # Check if interface is not None
                    if interface is not None:

                        # Continue if item is None
                        if item is None:
                            continue

                        # Add URL to item
                        item[_c.URL] = target.url

                        # Add timestamp to item
                        item[_c.YANKED_AT] = self.now()

                        # Convert to ORM item
                        item = interface.new(**item)

                        # Add anc commit item to database
                        self.db_session.add(item)
                        self.db_session.commit()

                        # Increment interface session count
                        interface.session_count += 1

            # Return the wrapped method
            return wrapped

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ WRAP METHODS                                                               │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Rename yank to yank start
        setattr(self, "yank_start", self.yank)

        # Get clean callback
        clean = getattr(self, "clean", None)

        # Check if clean is not null
        if clean:

            # Rename clean to clean start
            setattr(self, "clean_start", clean)

        # Get all methods
        all_methods = inspect.getmembers(self, predicate=inspect.ismethod)
        all_methods = {k: v for k, v in all_methods}

        # Get yank methods
        yank_methods = {k: v for k, v in all_methods.items() if k.startswith("yank_")}

        # Initialize has driver callback boolean
        has_driver_callback = False

        # Initialize has solve captcha callback boolean
        has_solve_captcha_callback = False

        # Iterate over yank methods
        for name, method in yank_methods.items():

            # Continue if suffixed method
            if "__" in name:
                continue

            # Get driver callback
            driver_callback = yank_methods.get(f"{name}__driver")

            # Check if driver callback is not null
            if driver_callback:

                # Set has driver callback to True
                has_driver_callback = True

            # Get clean callback
            clean_callback = all_methods.get(name.replace("yank_", "clean_", 1))

            # Get has captcha callback
            has_captcha_callback = yank_methods.get(f"{name}__has_captcha")

            # Get solve captcha callback
            solve_captcha_callback = yank_methods.get(f"{name}__solve_captcha")

            # Check if solve captcha callback is not null
            if solve_captcha_callback:

                # Set has has solve captcha callback to True
                has_solve_captcha_callback = True

            # Wrap and set yank method
            setattr(
                self,
                name,
                wrapper(
                    method,
                    driver_callback=driver_callback,
                    clean_callback=clean_callback,
                    has_captcha_callback=has_captcha_callback,
                    solve_captcha_callback=solve_captcha_callback,
                ),
            )

        # Rename _yank to yank
        setattr(self, "yank", self._yank)

        # Return callback booleans
        return has_driver_callback, has_solve_captcha_callback



# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ GENERAL IMPORTS                                                                    │
# └────────────────────────────────────────────────────────────────────────────────────┘

from datetime import datetime

# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ SQLALCHEMY IMPORTS                                                                 │
# └────────────────────────────────────────────────────────────────────────────────────┘

from sqlalchemy import Boolean, DateTime, exists, Float, func, Integer, String

# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ PROJECT IMPORTS                                                                    │
# └────────────────────────────────────────────────────────────────────────────────────┘

import yank.constants as _c


# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ INTERFACE DATABASE MIXIN                                                           │
# └────────────────────────────────────────────────────────────────────────────────────┘


class InterfaceDatabaseMixin:
    """ Interface Database Mixin """

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ CONSTANTS                                                                      │
    # └────────────────────────────────────────────────────────────────────────────────┘

    # Type Map
    TYPE_MAP = {
        str: String,
        int: Integer,
        float: Float,
        bool: Boolean,
        datetime: DateTime,
    }

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ CLASS ATTRIBUTES                                                               │
    # └────────────────────────────────────────────────────────────────────────────────┘

    # Initialize database session
    db_session = None

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ NEW                                                                            │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def new(self, **kwargs):
        """ A creates a new Item using the SQLAlchemy ORM Item class """

        # Get rid of all kwargs not in field map
        kwargs = {k: v for k, v in kwargs.items() if k in self.field_map}

        # Cast the item fields spplied as kawrgs and return an initialized Item object
        return self.Item(**self.cast_fields(kwargs))

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ GET                                                                            │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def get(self, queryset=None, **kwargs):
        """ Gets an item from the database based in the supplied kwargs """

        # Get queryset
        queryset = queryset or self.db_session.query(self.Item)

        # Return filtered queryset
        return queryset.filter_by(**kwargs).one()

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ COUNT                                                                          │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def count(self):
        """ Returns a count of items in the database """

        # Return count
        return self.db_session.query(func.count(self.Item.id)).scalar()

        """
        For more info on counting with SQL Alchemy:

        https://stackoverflow.com/questions/10822635/get-the-number-of-rows-in-table-
        using-sqlalchemy

        https://stackoverflow.com/questions/14754994/why-is-sqlalchemy-count-much-slower
        -than-the-raw-query
        """

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ EXISTS                                                                         │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def exists(self, **kwargs):
        """
        Returns a boolean of whether an item by the filter kwargs exists in the database
        """

        # Get Item
        Item = self.Item

        # Convert field names to field objects
        kwargs = {getattr(Item, field, None): value for field, value in kwargs.items()}

        # Remove null fields
        kwargs = {k: v for k, v in kwargs.items() if k}

        # Return boolean of whether or not the object exists
        return self.db_session.query(
            exists().where(*[k == v for k, v in kwargs.items()])
        ).scalar()

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ ALL                                                                            │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def all(self):
        """ Returns a queryset of all items in the database """

        # Return a queryset of all items
        return self.db_session.query(self.Item).all()

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ FILTER                                                                         │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def filter(self, queryset=None, display_map=None, tuples=None, **kwargs):
        """
        Returns a filtered queryset of items in the database by the provided kwargs
        """

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ TUPLES                                                                     │
        # └────────────────────────────────────────────────────────────────────────────┘

        # NOTE: A list of tuples can be used instead of kwargs for more complex logic
        # e.g. author__contains=a && author__contains=b
        # Using kwargs, there can only be one author__contains argument as it is a dict

        # Check if tuples
        if tuples:

            # Iterate over tuples
            for field, value in tuples:

                # Get queryset
                queryset = self.filter(
                    queryset=queryset, display_map=display_map, **{field: value}
                )

            # Return queryset
            return queryset

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ KWARGS                                                                     │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Get Item
        Item = self.Item

        # Get field map
        field_map = self.field_map

        # Initialize display map
        display_map = (
            {k.lower(): v for k, v in display_map.items()} if display_map else {}
        )

        # Get queryset
        queryset = queryset or self.db_session.query(Item)

        # Initialize field cache
        field_cache = {}

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ ITERATE OVER KWARGS                                                        │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Iterate over kwargs
        for field, value in kwargs.items():

            # Initialize should negate
            should_negate = False

            # Check if filed starts with tilde
            if field.startswith("~"):

                # Set should negate to True
                should_negate = True

                # Replace tilde
                field = field.replace("~", "", 1).strip()

            # Split field by modifier
            field_split = field.split("__")

            # Separate field from modifier
            field, modifier = (
                field_split if len(field_split) == 2 else (field_split[0], None)
            )

            # Set modifier
            modifier = modifier.lower() if modifier else ""

            # Check if field not in field map
            if field not in field_map:

                # Get field as display
                field = display_map.get(field.lower(), field)

            # Continue if field not in field map
            if field not in field_map:
                continue

            # Get field cast
            field_cast = field_map[field][_c.CAST]

            # Define field is string boolean
            field_is_str = field_cast is str

            # Add field object to field cache
            field_cache[field] = field_cache.get(field, getattr(Item, field))

            # Get field object
            field_obj = field_cache[field]

            # ┌────────────────────────────────────────────────────────────────────────┐
            # │ CAST VALUE                                                             │
            # └────────────────────────────────────────────────────────────────────────┘

            # Check if modifier is not a multiple value operator
            if modifier not in (_c.IN, _c.IIN):

                # Cast value
                value = self.cast_field(field, value)

            # ┌────────────────────────────────────────────────────────────────────────┐
            # │ SUBSTRING                                                              │
            # └────────────────────────────────────────────────────────────────────────┘

            # Handle case of substring
            if (
                modifier
                in (
                    _c.CONTAINS,
                    _c.ICONTAINS,
                    _c.STARTSWITH,
                    _c.ISTARTSWITH,
                    _c.ENDSWITH,
                    _c.IENDSWITH,
                )
                and field_is_str
            ):

                # ┌────────────────────────────────────────────────────────────────────┐
                # │ ARGUMENT STRING                                                    │
                # └────────────────────────────────────────────────────────────────────┘

                # Handle case of contains
                if modifier in (_c.CONTAINS, _c.ICONTAINS):

                    # Wrap argument string in % signs
                    arg_string = f"%{value}%"

                # Otherwise handle case of starts with
                elif modifier in (_c.STARTSWITH, _c.ISTARTSWITH):

                    # Add % sign to end of argument string
                    arg_string = f"{value}%"

                # Otherwise handle case of ends with
                elif modifier in (_c.ENDSWITH, _c.IENDSWITH):

                    # Add % sign to start of argument string
                    arg_string = f"%{value}"

                # ┌────────────────────────────────────────────────────────────────────┐
                # │ QUERY                                                              │
                # └────────────────────────────────────────────────────────────────────┘

                # Handle case of case-sensitive
                if modifier in (_c.CONTAINS, _c.STARTSWITH, _c.ENDSWITH):

                    # Set query using like
                    query = field_obj.like(arg_string)

                # Otherwise handle case of case-insensitive
                elif modifier in (_c.ICONTAINS, _c.ISTARTSWITH, _c.IENDSWITH):

                    # Set query using ilike
                    query = field_obj.ilike(arg_string)

            # ┌────────────────────────────────────────────────────────────────────────┐
            # │ REGEX                                                                  │
            # └────────────────────────────────────────────────────────────────────────┘

            # Otherwise handle case of regex
            elif modifier == _c.REGEX and field_is_str:

                # Define query
                query = field_obj.op("regexp")(value)

            # ┌────────────────────────────────────────────────────────────────────────┐
            # │ IN                                                                     │
            # └────────────────────────────────────────────────────────────────────────┘

            # Otherwise handle case of in
            elif modifier in (_c.IN, _c.IIN):

                # Convert and cast to a list of values
                values = [self.cast_field(field, i.strip()) for i in value.split(",")]

                # Handle case of iin
                if modifier == _c.IIN and field_is_str:

                    # Define query
                    query = func.lower(field_obj).in_([i.lower() for i in values])

                # Otherwise handle normal case
                else:

                    # Define query
                    query = field_obj.in_(values)

            # ┌────────────────────────────────────────────────────────────────────────┐
            # │ EXACT                                                                  │
            # └────────────────────────────────────────────────────────────────────────┘

            # Otherwise handle exact case
            else:

                # Check if modifier is iexact
                if modifier == _c.IEXACT and field_is_str:

                    # Define query
                    query = func.lower(field_obj) == value.lower()

                # Otherwise handle standard exact case
                else:

                    # Define query
                    query = field_obj == value

            # ┌────────────────────────────────────────────────────────────────────────┐
            # │ NEGATE                                                                 │
            # └────────────────────────────────────────────────────────────────────────┘

            # Check if should negate
            if should_negate:

                # Negagte query
                query = ~query

            # ┌────────────────────────────────────────────────────────────────────────┐
            # │ FILTER                                                                 │
            # └────────────────────────────────────────────────────────────────────────┘

            # Filter with query
            queryset = queryset.filter(query)

        # Return filtered queryset
        return queryset

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ SORT                                                                           │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def sort(self, *fields, queryset=None, display_map=None):
        """ Returns a queryset of items sorted by the provided fields """

        # Get Item
        Item = self.Item

        # Initialize display map
        display_map = (
            {k.lower(): v for k, v in display_map.items()} if display_map else {}
        )

        # Initialize new fields list
        _fields = []

        # Iterate over fields
        for field in fields:

            # Initialize ascending boolean
            ascending = True

            # Check if field begins with negative sign
            if field.startswith("-"):

                # Set ascending to False
                ascending = False

                # Remove negative sign from string
                field = field[1:]

            # Get SQL Alchemy field object by field name or display
            field = getattr(
                Item,
                field,
                getattr(Item, display_map[field.lower()], None)
                if field in display_map
                else None,
            )

            # Continue if field is None
            if field is None:
                continue

            # Apply sort direction to field object
            field = field.asc() if ascending is True else field.desc()

            # Add field to fields list
            _fields.append(field)

        # Get queryset
        queryset = queryset or self.db_session.query(Item)

        # Return sorted queryset
        return queryset.order_by(*_fields)



CAST = "cast"
CHROME = "chrome"
CHROMIUM = "chromium"
CONTAINS = "contains"
DISPLAY = "display"
EAGER = "eager"
ENDSWITH = "endswith"
FIREFOX = "firefox"
GET = "get"
HEADERS = "headers"
ICONTAINS = "icontains"
IENDSWITH = "iendswith"
IEXACT = "iexact"
ISTARTSWITH = "istartswith"
ID = "id"
IN = "in"
IIN = "iin"
INPUT_TAG = " <YNK:#> "
NONE = "none"
NORMAL = "normal"
NULL = "null"
QUICK = "quick"
RANK = "rank"
REGEX = "regex"
SESSION = "session"
STARTSWITH = "startswith"
TRANSIENT = "transient"
TYPE = "type"
UNIQUE = "unique"
URL = "url"
WEIGHT = "weight"
YANKED_AT = "yanked_at"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yank', 'yank.tools']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.11,<2.0.0',
 'arrow>=1.1.0,<2.0.0',
 'beautifulsoup4>=4.9.3,<5.0.0',
 'inflect>=5.3.0,<6.0.0',
 'lxml>=4.6.3,<5.0.0',
 'requests>=2.25.1,<3.0.0',
 'rich>=10.1.0,<11.0.0',
 'selenium-wire>=4.2.4,<5.0.0',
 'selenium>=3.141.0,<4.0.0',
 'tldextract>=3.1.0,<4.0.0',
 'undetected-chromedriver>=2.2.1,<3.0.0',
 'webdriver-manager>=3.4.0,<4.0.0',
 'xlwt>=1.3.0,<2.0.0']

setup_kwargs = {
    'name': 'yank',
    'version': '1.0.0',
    'description': 'A high-level web scraping utility kit for Python 3 and above',
    'long_description': None,
    'author': 'khunspoonzi',
    'author_email': 'khunspoonzi@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ GENERAL IMPORTS                                                                    │
# └────────────────────────────────────────────────────────────────────────────────────┘

import math
import re

# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ RICH IMPORTS                                                                       │
# └────────────────────────────────────────────────────────────────────────────────────┘

from rich.columns import Columns
from rich.console import Console, RenderGroup
from rich.padding import Padding
from rich.panel import Panel
from rich.table import Table

# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ PROJECT IMPORTS                                                                    │
# └────────────────────────────────────────────────────────────────────────────────────┘

import yank.constants as _c

from yank.tools import display_commands

# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ CONSTANTS                                                                          │
# └────────────────────────────────────────────────────────────────────────────────────┘

# Define AND operator
OPERATOR_AND = "&&"

# Define OR operator
OPERATOR_OR = "||"


# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ INTERFACE DISPLAY MIXIN                                                            │
# └────────────────────────────────────────────────────────────────────────────────────┘


class InterfaceDisplayMixin:
    """ Interface Display Mixin """

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ CLASS ATTRIBUTES                                                               │
    # └────────────────────────────────────────────────────────────────────────────────┘

    # Initialize cached console to None
    _console = None

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ CONSOLE                                                                        │
    # └────────────────────────────────────────────────────────────────────────────────┘

    @property
    def console(self):
        """ Returns a cached or newly initialized Rich console object """

        # Check if console is cached
        if self._console:

            # Return cached console
            return self._console

        # Initialize a new Rich console
        console = Console()

        # Cache console
        self._console = console

        # Return console
        return console

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ GET LIST RENDERABLE                                                            │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def get_list_renderable(
        self,
        interactive=False,
        limit=None,
        offset=0,
        sort_by=None,
        filter_by=None,
        highlight_id=None,
    ):
        """ Returns a Rich table renderable for the interface list view """

        # Define max limit
        limit_max = 1000

        # Initialize limie
        limit = limit or limit_max

        # Get limit where max is 1000
        limit = min(limit, limit_max)

        # Get display map
        display_map = self.list_display_map

        # Get items
        items = self.db_session.query(self.Item)

        # Check if filter by is not null
        if filter_by:

            # Apply filters
            items = self.filter(
                queryset=items, display_map=display_map, tuples=filter_by
            )

        # Check if sort by is not null
        if sort_by:

            # Apply sort fields
            items = self.sort(*sort_by, queryset=items, display_map=display_map)

        # Get row count
        row_count = items.count()

        # Get page count
        page_count = math.ceil(row_count / limit)

        # Define title
        title = (
            f"{self.name}: {self.db_table_name}\n"
            f"Page {int(offset / limit) + 1} of {page_count} "
            f"({row_count} {self.inflector.plural('row', row_count)})"
        )

        # Limit items
        items = items.offset(offset).limit(limit)

        # Initialize Rich Table as renderable
        renderable = Table(title=title)

        # Get display fields
        display_fields = self.display_list_by

        # Initialize fields
        fields = []

        # Iterate over display fields
        for field, display in display_fields.items():

            # Create column
            renderable.add_column(display)

            # Add field to fields
            fields.append(field)

        # Iterate over items
        for item in items:

            # Initialize style
            style = ""

            # Get item ID
            item_id = item.id

            # Check if item is highlighted
            if item_id == highlight_id:

                # Set style
                style = "white on blue"

            # Add row
            renderable.add_row(
                *[str(getattr(item, field)) for field in fields], style=style
            )

        # Check if interactive
        if interactive:

            # Initialize caption
            caption = "c\[ommands]: Show available commands"  # noqa

            # Initialize sub-caption
            sub_caption = ""

            # Check if sort by is not null
            if sort_by:

                # Add sort sub-caption
                sub_caption += f"sort {(' ' + OPERATOR_AND + ' ').join(sort_by)}\n"

            # Check if filter by is not null
            if filter_by:

                # Get filter strings
                filter_strings = [f"{f} = {v}" for f, v in filter_by]

                # Add filter sub-caption
                sub_caption += (
                    f"filter {(' ' + OPERATOR_AND + ' ').join(filter_strings)}\n"
                )

            # Check if sub-caption
            if sub_caption:

                # Combine caption and sub-caption
                caption = sub_caption + "\n" + caption

            # Set caption
            renderable.caption = caption.strip("\n")

        # Pad renderable
        renderable = Padding(renderable, (1, 1))

        # Return renderable and row count
        return renderable, row_count

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ DISPLAY LIST                                                                   │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def display_list(
        self,
        interactive=False,
        limit=20,
        offset=0,
        sort_by=None,
        filter_by=None,
    ):
        """ Displays a list of items using a Rich Table """

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ STATIC                                                                     │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Get renderable
        renderable, row_count = self.get_list_renderable(
            interactive=interactive,
            limit=limit,
            offset=offset,
            sort_by=sort_by,
            filter_by=filter_by,
        )

        # Get console
        console = self.console

        # Clear console
        console.clear()

        # Print renderable list view and return
        console.print(renderable, justify="center")

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ INTERACTIVE                                                                │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Return if not interactive
        if not interactive:
            return

        # Initialize highlight ID
        highlight_id = None

        # Define sort commands
        sort_commands = ("s", "sort")

        # Define filter commands
        filter_commands = ("f", "filter")

        # Define reset commands
        reset_commands = ("r", "reset")

        # Cache previous item URL
        item_url_previous = None

        # Initialize while loop
        while True:

            # Get command
            command = console.input(_c.INPUT_TAG)

            # ┌────────────────────────────────────────────────────────────────────────┐
            # │ BASIC COMMANDS                                                         │
            # └────────────────────────────────────────────────────────────────────────┘

            # Otherwise handle case of reset
            if command in reset_commands:

                # Reset sort fields
                sort_by = None

                # Reset filters
                filter_by = None

            # Otherwise handle case of quit
            elif command in ["q", "quit"]:
                break

            # Otherwise handle case of command
            elif command in ["c", "command", "commands"]:

                # Display commands
                display_commands(
                    "List View Commands",
                    console,
                    ("'", "<int> = 1", "Go to next nth page", "' 5"),
                    (";", "<int> = 1", "Go to previous nth page", "; 3"),
                    ("l[imit]", "<int>", "Set max number of rows per page", "l 10"),
                    (
                        "s[ort]",
                        f"<field>[ {OPERATOR_AND} <field>]",
                        "Sort rows by field(s)",
                        "s name && -age",
                    ),
                    (
                        "f[ilter]",
                        f"<field> = <value>[ {OPERATOR_AND} <field> = <value>]",
                        "Filter rows by field(s)",
                        "f name = Bob && age = 24",
                    ),
                    (
                        "r[eset]",
                        f"<command>[ {OPERATOR_AND} <command>] = sort && filter",
                        "Reset sort and / or filter parameters",
                        "r sort",
                    ),
                    ("d[etail]", "<id>", "Display a row in detail view", "d 5"),
                    ("o[pen]", "<id>", "Opens a row's URL in a browser", "o 5"),
                    aux_tables=[
                        (
                            "Filter Modifiers",
                            (
                                "modifier",
                                "usage",
                                "filters rows based on the...",
                                "example",
                            ),
                            (
                                (
                                    "exact",
                                    "<field> = <value>",
                                    "exact value of a field",
                                    "f name = Bob",
                                ),
                                (
                                    "iexact",
                                    "<field>__iexact = <value>",
                                    (
                                        "exact "
                                        "case-insensitive value of a string-based field"
                                    ),
                                    "f name = bob",
                                ),
                                (
                                    "contains",
                                    "<field>__contains = <substring>",
                                    (
                                        "presence of a "
                                        "substring in a string-based field"
                                    ),
                                    "f name__contains = Bo",
                                ),
                                (
                                    "icontains",
                                    "<field>__icontains = <substring>",
                                    (
                                        "presence of a "
                                        "case-insensitive substring in a string-based "
                                        "field"
                                    ),
                                    "f name__icontains = bo",
                                ),
                                (
                                    "startswith",
                                    "<field>__startswith = <substring>",
                                    (
                                        "presence of a "
                                        "substring at the start of a string-based "
                                        "field"
                                    ),
                                    "f name__startswith = Bo",
                                ),
                                (
                                    "istartswith",
                                    "<field>__istartswith = <substring>",
                                    (
                                        "presence of a "
                                        "case-insensitive substring at the start "
                                        "of a string-based field"
                                    ),
                                    "f name__istartswith = bo",
                                ),
                                (
                                    "endswith",
                                    "<field>__endswith = <substring>",
                                    (
                                        "presence of a "
                                        "substring at the end of a string-based "
                                        "field"
                                    ),
                                    "f name__endswith = ob",
                                ),
                                (
                                    "iendswith",
                                    "<field>__iendswith = <substring>",
                                    (
                                        "presence of a "
                                        "case-insensitive substring at the end "
                                        "of a string-based field"
                                    ),
                                    "f name__iendswith = OB",
                                ),
                                (
                                    "regex",
                                    "<field>__regex = <pattern>",
                                    (
                                        "positive match of a "
                                        "regex pattern against the value of a "
                                        "string-based field"
                                    ),
                                    "f name__regex = ^[A-Z]\[a-z]{2}$",  # noqa
                                ),
                                (
                                    "in",
                                    "<field>__in = <value>, <value>",
                                    (
                                        "exact value of any one of the "
                                        "options in a comma-separated list "
                                    ),
                                    "f name__in = Bob, Tom",
                                ),
                                (
                                    "iin",
                                    "<field>__iin = <value>, <value>",
                                    (
                                        "exact case-insensitive value of a "
                                        "string-based field against any one of the "
                                        "options in a comma-separated list "
                                    ),
                                    "f name__iin = bob, tom",
                                ),
                            ),
                        ),
                        (
                            "Filter Operators",
                            (
                                "operator",
                                "usage",
                                "description",
                                "example",
                            ),
                            (
                                (
                                    "~",
                                    "~ <field>[__modifier] = <value>",
                                    (
                                        "Negates the filter to exclude the selected "
                                        "rows"
                                    ),
                                    "f ~ name = Bob",
                                ),
                                (
                                    "&&",
                                    (
                                        "<field>[__modifier] = <value> && "
                                        "<field>[__modifier] = <value>"
                                    ),
                                    (
                                        "Selects rows that satisfy all of the chained "
                                        "filters"
                                    ),
                                    "f name = Bob && age = 24",
                                ),
                            ),
                        ),
                    ],
                )

            # ┌────────────────────────────────────────────────────────────────────────┐
            # │ ARGUMENT COMMANDS                                                      │
            # └────────────────────────────────────────────────────────────────────────┘

            # Otherwise handle more complex commands
            else:

                # Get command and index
                match = re.search(r"([a-zA-Z';]+) *(.*)", command)

                # Separate args from command
                command, args = (
                    (match.group(1).strip(), match.group(2)) if match else (None, None)
                )

                # ┌────────────────────────────────────────────────────────────────────┐
                # │ NEXT PAGE                                                          │
                # └────────────────────────────────────────────────────────────────────┘

                # Handle case of next page
                if command == "'":

                    # Define multiplier
                    multiplier = int(args) if args.isdigit() else 1

                    # Get new offset
                    new_offset = offset + (limit * multiplier)

                    # Increment offset by limit
                    offset = offset if new_offset >= row_count else new_offset

                # ┌────────────────────────────────────────────────────────────────────┐
                # │ PREVIOUS PAGE                                                      │
                # └────────────────────────────────────────────────────────────────────┘

                # Otherwise handle case of previous page
                elif command == ";":

                    # Define multiplier
                    multiplier = int(args) if args.isdigit() else 1

                    # Get new offset
                    new_offset = offset - (limit * multiplier)

                    # Decrement offset by limit
                    offset = max(new_offset, 0)

                # ┌────────────────────────────────────────────────────────────────────┐
                # │ LIMIT                                                              │
                # └────────────────────────────────────────────────────────────────────┘

                # Otherwise handle case of limit
                elif command in ["l", "limit"]:

                    # Set new limit
                    limit = max(int(args), 0) if args.isdigit() else None

                    # Reset offset to 0
                    offset = 0

                # ┌────────────────────────────────────────────────────────────────────┐
                # │ SORT                                                               │
                # └────────────────────────────────────────────────────────────────────┘

                # Otherwise handle case of sort
                elif command in sort_commands:

                    # Set sort by argument
                    sort_by = [f.strip() for f in args.split(OPERATOR_AND)]

                    # Remove null items
                    sort_by = [s for s in sort_by if s]

                    # Set to None if null
                    sort_by = sort_by or None

                    # Reset offset to zero
                    offset = 0

                # ┌────────────────────────────────────────────────────────────────────┐
                # │ FILTER                                                             │
                # └────────────────────────────────────────────────────────────────────┘

                # Otherwise handle case of filter
                elif command in filter_commands:

                    # Set filter by argument
                    filter_by = [f.strip() for f in args.split(OPERATOR_AND)]

                    # Convert filter by to list of tuples
                    filter_by = (
                        [[f.strip() for f in arg.split("=")] for arg in filter_by]
                        if filter_by
                        else None
                    )

                    # Reset offset to zero
                    offset = 0

                # ┌────────────────────────────────────────────────────────────────────┐
                # │ RESET                                                              │
                # └────────────────────────────────────────────────────────────────────┘

                # Otherwise handle case of reset
                elif command in reset_commands:

                    # Iterate over split args
                    for arg in (args or "").split(OPERATOR_AND):

                        # Strip arg
                        arg = arg.strip()

                        # Handle case of sort
                        if arg in sort_commands:

                            # Set sort by to None
                            sort_by = None

                        # Otherwise handle case of filter
                        elif arg in filter_commands:

                            # Set filter by to None
                            filter_by = None

                # ┌────────────────────────────────────────────────────────────────────┐
                # │ DETAIL                                                             │
                # └────────────────────────────────────────────────────────────────────┘

                # Otherwise handle case of detail
                elif command in ["d", "detail"]:

                    # Get item ID
                    item_id = int(args)

                    # Display interface detail view
                    self.display_detail(item_id=item_id, interactive=interactive)

                # ┌────────────────────────────────────────────────────────────────────┐
                # │ OPEN                                                               │
                # └────────────────────────────────────────────────────────────────────┘

                # Otherwise handle case of open
                elif command in ["o", "open"]:

                    # Get item ID
                    item_id = int(args)

                    # Set highlight ID
                    highlight_id = item_id

                    # Get item
                    item = self.get(id=item_id)

                    # Get item URL
                    item_url = item.url

                    # Check if item URL does not equal previous item URL
                    if item_url != item_url_previous:

                        # Get item URL with driver
                        self.browser.driver.get(item_url)

                    # Cache previous item URL
                    item_url_previous = item_url

            # ┌────────────────────────────────────────────────────────────────────────┐
            # │ RE-RENDER                                                              │
            # └────────────────────────────────────────────────────────────────────────┘

            # Get renderable and row count
            renderable, row_count = self.get_list_renderable(
                interactive=interactive,
                limit=limit,
                offset=offset,
                sort_by=sort_by,
                filter_by=filter_by,
                highlight_id=highlight_id,
            )

            # Clear console
            console.clear()

            # Print renderable list view and return
            console.print(renderable, justify="center")

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ GET DETAIL RENDERABLE                                                          │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def get_detail_renderable(self, item_id):

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ GET ITEM                                                                   │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Get item by ID
        item = self.get(id=int(item_id))

        # Define panel kwargs
        panel_kwargs = {"title_align": "right"}

        # Get display detail by
        display_detail_by = self.display_detail_by

        # Initialize columns
        columns = []

        # Iterate over rows in display detail by
        for row in display_detail_by:

            # Initialize columns
            cols = []

            # Iterate over columns in row
            for field, display in row.items():

                # Get value
                value = getattr(item, field, "N/A")

                # Stringify value
                value = str(value)

                # Add Panel to columns
                cols.append(Panel(value, title=display, **panel_kwargs))

            # Append row of columns
            columns.append(Columns(cols, expand=True))

        # Create renderable
        renderable = RenderGroup(*columns)

        # Pad renderable
        renderable = Padding(renderable, (1, 1))

        # Return renderable
        return renderable

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ DISPLAY DETAIL                                                                 │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def display_detail(self, item_id, interactive=False):
        """ Displays an item detail view using Rich Panels """

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ STATIC                                                                     │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Get detail renderable
        renderable = self.get_detail_renderable(item_id=item_id)

        # Get console
        console = self.console

        # Clear console
        console.clear()

        # Print renderable list view and return
        console.print(renderable, justify="center")

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ INTERACTIVE                                                                │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Return if not interactive
        if not interactive:
            return

        # Initialize while loop
        while True:

            # Get command
            command = console.input(_c.INPUT_TAG)

            # ┌────────────────────────────────────────────────────────────────────────┐
            # │ BASIC COMMANDS                                                         │
            # └────────────────────────────────────────────────────────────────────────┘

            # Handle case of quit
            if command in ["q", "quit"]:
                break

            # ┌────────────────────────────────────────────────────────────────────────┐
            # │ RE-RENDER                                                              │
            # └────────────────────────────────────────────────────────────────────────┘

            # Clear console
            console.clear()

            # Print renderable list view and return
            console.print(renderable, justify="center")



# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ GENERAL IMPORTS                                                                    │
# └────────────────────────────────────────────────────────────────────────────────────┘

import math

from datetime import datetime

# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ SQLALCHEMY IMPORTS                                                                 │
# └────────────────────────────────────────────────────────────────────────────────────┘

from sqlalchemy import Column, func, Integer
from sqlalchemy.ext.hybrid import hybrid_property

# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ PROJECT IMPORTS                                                                    │
# └────────────────────────────────────────────────────────────────────────────────────┘

import yank.constants as _c

from yank.browser import Browser
from yank.interface_database_mixin import InterfaceDatabaseMixin
from yank.interface_display_mixin import InterfaceDisplayMixin


# ┌────────────────────────────────────────────────────────────────────────────────────┐
# │ INTERFACE                                                                          │
# └────────────────────────────────────────────────────────────────────────────────────┘


class Interface(InterfaceDatabaseMixin, InterfaceDisplayMixin):
    """ A utility class for managing the schema of yanked data """

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ CONSTANTS                                                                      │
    # └────────────────────────────────────────────────────────────────────────────────┘

    # Interface
    CAST = _c.CAST
    DISPLAY = _c.DISPLAY
    NULL = _c.NULL
    RANK = _c.RANK
    UNIQUE = _c.UNIQUE
    WEIGHT = _c.WEIGHT

    # Fields
    ID = _c.ID
    URL = _c.URL
    YANKED_AT = _c.YANKED_AT

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ CLASS ATTRIBUTES                                                               │
    # └────────────────────────────────────────────────────────────────────────────────┘

    # Initialize name to empty string
    name = ""

    # Initialize skip by URL to False
    skip_by_url = False

    # Initialize session limit to None
    session_limit = None

    # Initialize session count to 0
    session_count = 0

    # Initialize display list by to None
    display_list_by = None

    # Initialize list display map to None
    list_display_map = None

    # Initialize display detail by to None
    display_detail_by = None

    # Initialize detail display map to None
    detail_display_map = None

    # Initialize default browser
    default_browser = Browser.CHROME

    # Initialize cached browser
    _browser = None

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ INIT METHOD                                                                    │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def __init__(self, DBBase, db_table_name, inflector, **kwargs):
        """ Init Method """

        # Get common constants
        CAST = self.CAST
        DISPLAY = self.DISPLAY
        RANK = self.RANK
        UNIQUE = self.UNIQUE
        WEIGHT = self.WEIGHT

        # Set table name
        self.db_table_name = db_table_name

        # Set inflector
        self.inflector = inflector

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ CUSTOM ATTRIBUTES                                                          │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Get attributes
        attributes = {k: v for k, v in kwargs.items() if k.startswith("__")}

        # Iterate over attributes
        for k, v in attributes.items():

            # Set attribute
            setattr(self, k.replace("__", "", 1).strip(), v)

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ FIELD MAP                                                                  │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Initialize field map
        field_map = {_c.ID: {_c.CAST: int, _c.DISPLAY: "ID"}}

        # Add kwargs to field map
        field_map = {
            **field_map,
            **{k: v for k, v in kwargs.items() if k not in attributes},
        }

        # Add yanked at and URL to field map
        field_map[self.URL] = {CAST: str, DISPLAY: "URL"}
        field_map[self.YANKED_AT] = datetime

        # Normalize the structure of the field map
        field_map = {
            k: (v if type(v) is dict else {CAST: v}) for k, v in field_map.items()
        }

        # Set field map
        self.field_map = field_map

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ DEFAULTS                                                                   │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Compute total weight
        weight_total = sum(
            [abs(info.get(WEIGHT, 0)) for info in self.field_map.values()]
        )

        # Iterate over field map
        for field, info in self.field_map.items():

            # Define display
            info[DISPLAY] = info.get(DISPLAY) or " ".join(
                [w.title() for w in field.split("_")]
            )

            # Get weight
            weight = info.get(WEIGHT) or None

            # Determine if weight is reversed
            weight_is_reversed = weight < 0 if weight else False

            # Normalize weight according to total
            weight = abs(weight) / weight_total if weight else None

            # Reverse weight if necessary
            weight = weight * -1 if weight_is_reversed else weight

            # Set weight
            info[WEIGHT] = weight

        # Get weighted columns
        weighted_columns = {
            field: info[WEIGHT]
            for field, info in self.field_map.items()
            if info[WEIGHT] is not None
        }

        # Get positive weighted columns
        weighted_columns_positive = {k: v for k, v in weighted_columns.items() if v > 0}

        # Get negative weighted columns
        weighted_columns_negative = {k: v for k, v in weighted_columns.items() if v < 0}

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ COLUMNS                                                                    │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Define columns decorator
        def columns(cls):
            """ Dynamically sets SQLAlchemy columns on a target ORM class """

            # Iterate over field map
            for field, info in self.field_map.items():

                # Continue if field is ID (created manually)
                if field == _c.ID:
                    continue

                # Get column type
                ColType = self.TYPE_MAP[info[CAST]]

                # Get unique
                unique = info.get(UNIQUE, False)

                # Set class attribute
                setattr(cls, field, Column(ColType, unique=unique))

            # Define rank
            @hybrid_property
            def rank(self):

                # Get rank
                rank = math.prod(
                    [1]
                    + [
                        getattr(self, field) ** weight
                        for field, weight in weighted_columns_positive.items()
                    ]
                ) / math.prod(
                    [1]
                    + [
                        getattr(self, field) ** abs(weight)
                        for field, weight in weighted_columns_negative.items()
                    ]
                )

                # Return rank
                return rank

            # Define rank expression
            @rank.expression
            def rank(cls):

                # Get rank
                rank = math.prod(
                    [1]
                    + [
                        func.pow(getattr(cls, field), weight)
                        for field, weight in weighted_columns_positive.items()
                    ]
                ) / math.prod(
                    [1]
                    + [
                        func.pow(getattr(cls, field), abs(weight))
                        for field, weight in weighted_columns_negative.items()
                    ]
                )
                return rank

            # Set rank property
            setattr(cls, RANK, rank)

            # Return the ORM class
            return cls

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ ITEM                                                                       │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Define Item ORM class
        @columns
        class Item(DBBase):
            """ The ORM class that will house the user-defined columns """

            # Set table name
            __tablename__ = self.db_table_name

            # Set ID as primary key
            id = Column(Integer, primary_key=True)

        # Set Item class on interface object
        self.Item = Item

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ LIST DISPLAY SETTINGS                                                      │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Initialize list display map
        self.list_display_map = {self.field_map[_c.ID][_c.DISPLAY]: _c.ID}

        # Check if display list by is defined
        if self.display_list_by:

            # Iterate over display list by
            for item in self.display_list_by:

                # Initialize display
                display = None

                # Check if item is a list or tuple
                if type(item) in (list, tuple):

                    # Unpack field and display
                    field, display = item

                # Otherwise handle string
                else:

                    # Set field
                    field = item

                # Continue if field not in field map
                if field not in self.field_map:
                    continue

                # Set display
                display = display or self.field_map[field][_c.DISPLAY]

                # Add to display map
                self.list_display_map[display] = field

        # Otherwise handle case of undefined display list by
        else:

            # Iterate over the first five items of the field map
            for field in list(self.field_map.keys())[:5]:

                # Add to display map
                self.list_display_map[self.field_map[field][_c.DISPLAY]] = field

        # Convert display list by into a dictionary
        self.display_list_by = {v: k for k, v in self.list_display_map.items()}

        # ┌────────────────────────────────────────────────────────────────────────────┐
        # │ DETAIL DISPLAY SETTINGS                                                    │
        # └────────────────────────────────────────────────────────────────────────────┘

        # Initialize detail display map
        self.detail_display_map = {}

        # Initialize new display detail by
        _display_detail_by = []

        # Check if display detail by is defined
        if self.display_detail_by:

            # Iterate over rows in display detail by
            for row in self.display_detail_by:

                # Initialize columns
                cols = {}

                # Iterate over columns in row
                for col in row:

                    # Initialize display
                    display = None

                    # Check if item is a list or tuple
                    if type(col) in (list, tuple):

                        # Unpack field and display
                        field, display = col

                    # Otherwise handle string
                    else:

                        # Set field
                        field = col

                    # Continue if field not in field map
                    if field not in self.field_map:
                        continue

                    # Set display
                    display = display or self.field_map[field][_c.DISPLAY]

                    # Add column to columns
                    cols[field] = display

                    # Add to display map
                    self.detail_display_map[display] = field

                # Add columns to new display detail by
                _display_detail_by.append(cols)

        # Otherwise handle case of undefined display detail by
        else:

            # Initialize columns
            cols = {}

            # Get detail display fields as first ten fields
            detail_display_fields = list(self.field_map.keys())[:10]

            # Remove URL and yanked at from fields
            detail_display_fields = [
                f for f in detail_display_fields if f not in (_c.URL, _c.YANKED_AT)
            ]

            # Iterate over detail display fields
            for i, field in enumerate(detail_display_fields, 1):

                # Get display
                display = self.field_map[field][_c.DISPLAY]

                # Add column to columns
                cols[field] = display

                # Add to display map
                self.detail_display_map[display] = field

                # Check if index is divisible by two or is last
                if i % 2 == 0 or i == len(detail_display_fields):

                    # Add columns to new display detail by
                    _display_detail_by.append(cols)

                    # Reset columns
                    cols = {}

            # Get URL display
            url_display = self.field_map[_c.URL][_c.DISPLAY]

            # Get yanked at display
            yanked_at_display = self.field_map[_c.YANKED_AT][_c.DISPLAY]

            # Add to display map
            self.detail_display_map[url_display] = _c.URL
            self.detail_display_map[yanked_at_display] = _c.YANKED_AT

            # Add URL and yanked at as a pair at the end
            _display_detail_by.append(
                {_c.URL: url_display, _c.YANKED_AT: yanked_at_display}
            )

        # Redefine display detail by
        self.display_detail_by = _display_detail_by

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ BROWSER                                                                        │
    # └────────────────────────────────────────────────────────────────────────────────┘

    @property
    def browser(self):
        """ Returns a cached browser instance """

        # Check if browser is cached
        if self._browser:

            # Return cached browser
            return self._browser

        # Initialize a new browser
        browser = Browser(
            self.default_browser,
            driver_mode=Browser.NORMAL,
            driver_headless=False,
            driver_requests=False,
        )

        # Cache browser
        self._browser = browser

        # Return browser
        return browser

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ CAST FIELDS                                                                    │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def cast_fields(self, item_dict, display_map=None):
        """ Casts a dict of item fields according to a the interface's field map """

        # Initialize display map
        display_map = (
            {k.lower(): v for k, v in display_map.items()} if display_map else {}
        )

        # Convert fields from display
        item_dict = {
            field
            if field in self.field_map
            else display_map.get(field.lower(), field): value
            for field, value in item_dict.items()
        }

        # Cast item dict
        item_dict = {
            field: self.cast_field(field, value) for field, value in item_dict.items()
        }

        # Return the item dict
        return item_dict

    # ┌────────────────────────────────────────────────────────────────────────────────┐
    # │ CAST FIELD                                                                     │
    # └────────────────────────────────────────────────────────────────────────────────┘

    def cast_field(self, field, value):
        """ Casts a value according to a field in the interface's field map """

        # Get field map
        field_map = self.field_map

        # Return value if field not in field map
        if field not in field_map:
            return value

        # Check if value is None
        if value is None:

            # Get null boolean
            null = field_map[field].get(self.NULL, False)

            # Check if null is permitted
            if null:

                # Return the value
                return value

        # Get value type
        value_type = type(value)

        # Check if value is a list
        if value_type is list:

            # Return a comma separated list
            # Currently no support for array structures in database
            return "; ".join([str(i).replace(";", ",") for i in value])

        # Get to type
        to_type = field_map[field][self.CAST]

        # Check if to type is int
        if to_type is int:

            # Check if value is str
            if value_type is str:

                # Remove common unwanted characters
                value = value.replace(",", "")

        # Handle case of datetime
        elif to_type is datetime:

            # Assert value is datetime
            assert value_type is datetime, f"{field} is not a valid datetime"

        # Cast value to appropriate type
        value = to_type(value)

        # Return value
        return value
